/* Hello, Chord converter
 *********
 * TODO: punctuation stuff (eliminate stuff like .; or ,;...) 
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <ctype.h>
#include <assert.h>
#include <zlib.h>

#include "common.h"
typedef int DWord;
typedef short Word;

#define dmDBNameLength    32/* 31 chars + 1 null terminator */
#define MAXLINELEN	4096
#define MAC_MINUS_UNIX_TIME (2082844800L)

#pragma pack(1)
struct pdb_header {       /* 78 bytes total */
     char   name[ dmDBNameLength ];
     Word   attributes;
     Word   version;
     DWord  create_time;
     DWord  modify_time;
     DWord  backup_time;
     DWord  modificationNumber;
     DWord  appInfoID;
     DWord  sortInfoID;
     char   type[4];
     char   creator[4];
     DWord  id_seed;
     DWord  nextRecordList;
     Word   numRecords;
};

struct pdb_rec_header {   /* 8 bytes total */
     DWord  offset;
     struct {
            int delete    : 1;
            int dirty     : 1;
            int busy      : 1;
            int secret    : 1;
            int category  : 4;
     }      attributes;
     char   uniqueID[3];
};

void *zlib_alloc(void *opaque, uInt items, uInt size) {
	return malloc(items * size);
}

void zlib_free(void *opaque, void *address) {
	free(address);
}

void zInit(z_stream *z) {
	z->avail_in = z->total_in = z->avail_out = z->total_out = 0;
	z->zalloc = zlib_alloc;
	z->zfree = zlib_free;
	z->opaque = NULL;
	deflateInit(z, Z_BEST_COMPRESSION);
}

void zFinalFlushBuffer(z_stream *z, char *zbuf, int zbuflen, FILE *out) {
	z->next_out = zbuf;
	z->avail_out = zbuflen;
	while (deflate(z, Z_FINISH) != Z_STREAM_END) {
		fwrite(zbuf, zbuflen - z->avail_out, 1, out);
		z->next_out = zbuf;
		z->avail_out = zbuflen;
	}
	fwrite(zbuf, zbuflen - z->avail_out, 1, out);
}

void zFlushBuffer(z_stream *z, char *zbuf, int zbuflen, FILE *out) {
	while (z->avail_in) {
		z->next_out = zbuf;
		z->avail_out = zbuflen;
		if (deflate(z, Z_NO_FLUSH) != Z_OK)
			printf("Problems...\n");
		fwrite(zbuf, zbuflen - z->avail_out, 1, out);
	}
}

void zCompressBuffer(z_stream *z, const char *buf, char *zbuf, int zbuflen,
		     FILE *out) {
	z->next_in  = (char *)buf; /* trust zlib not to do bad stuff */
	z->avail_in = strlen(buf);
	zFlushBuffer(z, zbuf, zbuflen, out);
}

void headerInit(struct pdb_header *ph, int recs) {
	struct timeval tv;
	struct timezone tz;

	gettimeofday(&tv, &tz);
	bzero(ph, sizeof(*ph));
	strncpy(ph->name, dbName, dmDBNameLength-1);
	ph->attributes = htons(0);
	ph->version = htons(1);
	ph->create_time = htonl(tv.tv_sec + MAC_MINUS_UNIX_TIME);
	ph->modify_time = ph->create_time;
	ph->backup_time = ph->create_time;
	ph->modificationNumber = 0;
	ph->appInfoID = 0;
	ph->sortInfoID = 0;
	strncpy(ph->type, dbType_s, 4);
	strncpy(ph->creator, appCreatorID_s, 4);
	ph->id_seed = 0;
	ph->nextRecordList = 0;
	ph->numRecords = ntohs(recs);
}

int main(int argc, char **argv) {
	FILE   *in;
	FILE   *out;
	long    sec;
	long    loc,oldloc;
	short   origlen;
	int     i;
	int     first;
	char    buf[MAXLINELEN];
	char    zbuf[MAXLINELEN];
	char   *ptr;
	char   *ptr2;
	z_stream z;
	struct pdb_header ph;
	struct pdb_rec_header prh;

	if (argc <= 2) {
		fprintf(stderr, "Usage: %s outfile infile1 ...\n", *argv);
		exit(-1);
	}

	out = fopen(argv[1], "w");
	loc = sizeof(ph) + (argc-2)*sizeof(prh);

	headerInit(&ph, argc-2);
	bzero(&prh, sizeof(prh));

	fwrite(&ph, sizeof(ph), 1, out);
	for (i=0; i<argc-2; i++)
		fwrite(&prh, sizeof(prh), 1, out);

	for (i=0; i<argc-2; i++) {
		/* fix the header to have correct offset */
		fseek(out, sizeof(ph) + i*sizeof(prh), SEEK_SET);
		prh.offset = htonl(loc);
		fwrite(&prh, sizeof(prh), 1, out);
		fseek(out, 0, SEEK_END);

		zInit(&z);
		z.next_out = zbuf;
		z.next_in  = buf;
		fputs("  ", out); /* filled in later with final length */

		/* parse the filename and put it somewhere */
		in = fopen(argv[i+2], "r");
		strncpy(buf, argv[i+2], MAXLINELEN);
		if (ptr = strrchr(buf, '.'))
			*ptr = 0;
		while (ptr = strchr(buf, '_'))
			*ptr = ' ';
		zCompressBuffer(&z, buf, zbuf, MAXLINELEN, out);
		zCompressBuffer(&z, ":", zbuf, MAXLINELEN, out);

		first = 1;

		while (fgets(buf, MAXLINELEN, in)) {
			/* s/\{[^}]*\}//g; */
			while (ptr = strchr(buf, '{')) {
				if (ptr2 = strchr(ptr, '}'))
					memmove(ptr, ptr2+1, 
						(buf+MAXLINELEN) - (ptr2+1));
				else
					*ptr = 0;
			}

			/* s/\s+$//; */
			while (*buf && isspace(buf[strlen(buf)-1]))
				buf[strlen(buf)-1] = 0;

			if (*buf) {
				if (first) {
					first = 0;
				} else {
					zCompressBuffer(&z, "; ", zbuf, 
							MAXLINELEN, out);
				}
				zCompressBuffer(&z, buf, zbuf, MAXLINELEN, out);
			}
		}

		/* don't really need this stuff; take it out after everything
		 * works with it in. ych 2/26/01
		 */
		*buf = 0;
		z.next_in = buf;
		z.avail_in = 1;
		zFinalFlushBuffer(&z, zbuf, MAXLINELEN, out);

		oldloc = loc;
		loc += 2+z.total_out;
		origlen = htons((short)z.total_in);
		fseek(out, oldloc, SEEK_SET);
		fwrite(&origlen, sizeof(short), 1, out);

		deflateEnd(&z);
		fclose(in);
	}
	fclose(out);
}
